#include <string.h>
#include <stdlib.h>
#include "tic.h"

static const char FRAME_START = 0x02;
static const char GROUP_START = 0x0a;
static const char FIELD_SEP = 0x20;

DecodeState* create_decode_state(size_t queue_size) {
    DecodeState* out = malloc(sizeof(DecodeState));
    out->cur_frame = NULL;
    out->frame_queue = xQueueCreate(queue_size, sizeof(InfoFrame*));
    out->checksum = 0;
    out->fill_step = GFS_NotStarted;
    out->postpone_buffer[0] = '\0';
    out->postpone_buffer_size = 0;
    return out;
}

void free_decode_state(DecodeState* st) {
    vQueueDelete(st->frame_queue);
    if(st->cur_frame != NULL) {
        tic_delete_frame(st->cur_frame);
        free(st->cur_frame);
    }
    free(st);
}

void tic_delete_group(InfoGroup* group) {
    free(group->time_str);
    if(group->data_type == GROUP_DATA_STR)
        free(group->data.str);
}
void tic_delete_frame(InfoFrame* frame) {
    for(size_t g_id=0; g_id < frame->group_count; ++g_id)
        tic_delete_group(frame->groups + g_id);
}

char* tic_discard_until_frame(char* buffer, size_t buf_size) {
    size_t pos=0;
    for(pos=0; pos < buf_size; ++pos) {
        if(buffer[pos] == FRAME_START) {
            break;
        }
    }
    if(pos == buf_size)
        return NULL;
    return buffer + pos;
}

const uint32_t BAD_STR_TO_INT = ~0U;
uint32_t str_to_int(char* buffer, const char* buffer_end) {
    uint64_t res = 0;
    for(char* chr=buffer; chr != buffer_end; ++chr) {
        if('0' <= *chr && *chr <= '9') {
            res *= 10;
            res += (*chr - '0');
        } else {
            return BAD_STR_TO_INT;
        }
    }
    if(res > BAD_STR_TO_INT)
        return BAD_STR_TO_INT;
    return res;
}

static void get_data_pointers(char** start, char** end, const char* buffer, char* pos, DecodeState* state) {
    if(*start == NULL) { // Postponed data
        *start = state->postpone_buffer;
        size_t copy_size = pos - buffer;
        if(copy_size + state->postpone_buffer_size > POSTPONE_BUFFER_MAXSIZE)
            copy_size = POSTPONE_BUFFER_MAXSIZE - state->postpone_buffer_size;
        char* postpone_end = state->postpone_buffer
            + state->postpone_buffer_size;
        memcpy(
            postpone_end,
            buffer,
            copy_size);
        *end = postpone_end + copy_size;
    } else { // No postponed data
        *end = pos;
    }
}

int tic_decode(char* buffer, size_t buf_size, DecodeState* state) {
    int frames_decoded = 0;
    char* start = NULL;
    char* data_end = NULL;

    for(char* pos = buffer; pos != buffer + buf_size; ++pos) {
        if(state->fill_step == GFS_Checksum) { // End of frame -- verify checksum
            if(((state->checksum & 0x3F) + 0x20) != *pos) {
                state->fill_step = GFS_Invalid;
                continue;
            }
            state->fill_step = GFS_Done;
            continue;
        }
        state->checksum += *pos; // Update checksum

        if(*pos == FRAME_START) {
            if(state->cur_frame != NULL) {
                frames_decoded += 1;
                if(uxQueueSpacesAvailable(state->frame_queue) == 0) {
                    tic_delete_frame(state->cur_frame);
                    free(state->cur_frame);
                }
                else
                    xQueueSend(state->frame_queue, state->cur_frame, 0);
            }
            state->cur_frame = malloc(sizeof(InfoFrame));
            memset(state->cur_frame, 0, sizeof(InfoFrame));
            state->cur_frame->group_count = 0;
        }
        else if(*pos == GROUP_START) {
            // Reset checksum
            state->checksum = 0;

            if(state->fill_step == GFS_Done) {
                state->cur_frame->group_count++;
                state->fill_step = GFS_NotStarted;
            } else if(state->fill_step == GFS_Invalid) {
                memset(state->cur_frame->groups + state->cur_frame->group_count,
                        '0', sizeof(InfoGroup));
                state->fill_step = GFS_NotStarted;
            } else if(state->fill_step == GFS_NotStarted) {
                state->fill_step = GFS_Tag;
                start = pos + 1;
            } else {
                state->fill_step = GFS_Invalid;
            }
        }
        else if(*pos == FIELD_SEP) {
            InfoGroup* cur_group =
                state->cur_frame->groups + state->cur_frame->group_count;
            if(state->fill_step == GFS_Tag) {
                get_data_pointers(&start, &data_end, buffer, pos, state);
                size_t label_size = data_end - start + 1;
                if(label_size > INFO_GROUP_LABEL_MAX_SIZE)
                    label_size = INFO_GROUP_LABEL_MAX_SIZE;
                memcpy(cur_group->label, start, label_size-1);
                cur_group->label[label_size-1] = '\0';
                state->fill_step = GFS_Data;
                start = pos + 1;
            }
            else if(state->fill_step == GFS_Data) {
                get_data_pointers(&start, &data_end, buffer, pos, state);
                uint32_t int_val = str_to_int(start, data_end);
                if(int_val == BAD_STR_TO_INT) {
                    size_t len = data_end - start + 1;
                    cur_group->data_type = GROUP_DATA_STR;
                    cur_group->data.str = malloc(sizeof(char) * len);
                    memcpy(cur_group->data.str, start, len-1);
                    cur_group->data.str[len-1] = '\0';
                } else {
                    cur_group->data_type = GROUP_DATA_NUM;
                    cur_group->data.num = int_val;
                }
                state->fill_step = GFS_Checksum;
                state->checksum -= *pos;
            }
        }
    }

    if((state->fill_step == GFS_Tag || state->fill_step == GFS_Data) && start != NULL) {
        size_t postpone_size = (buffer + buf_size) - start;
        if(postpone_size > POSTPONE_BUFFER_MAXSIZE)
            postpone_size = POSTPONE_BUFFER_MAXSIZE;
        memcpy(state->postpone_buffer, start, postpone_size);
        state->postpone_buffer_size = postpone_size;
    }

    return frames_decoded;
}
